package consoledoctor.barcode_scanner

import android.app.Activity
import android.net.Uri
import android.os.Handler
import android.os.Looper
import android.util.Size
import androidx.camera.core.CameraSelector
import androidx.camera.core.ExperimentalGetImage
import com.google.mlkit.vision.barcode.BarcodeScannerOptions
import consoledoctor.barcode_scanner.objects.BarcodeFormats
import consoledoctor.barcode_scanner.objects.DetectionSpeed
import consoledoctor.barcode_scanner.objects.BarcodeScannerErrorCodes
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.PluginRegistry.RequestPermissionsResultListener
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.view.TextureRegistry
import java.io.File

class BarcodeScannerHandler(
    private val activity: Activity,
    private val barcodeHandler: BarcodeHandler,
    binaryMessenger: BinaryMessenger,
    private val permissions: BarcodeScannerPermissions,
    private val addPermissionListener: (RequestPermissionsResultListener) -> Unit,
    textureRegistry: TextureRegistry): MethodChannel.MethodCallHandler {

    private val analyzeImageErrorCallback: AnalyzerErrorCallback = {
        Handler(Looper.getMainLooper()).post {
            analyzerResult?.error(BarcodeScannerErrorCodes.BARCODE_ERROR, it, null)
            analyzerResult = null
        }
    }

    private val analyzeImageSuccessCallback: AnalyzerSuccessCallback = {
        Handler(Looper.getMainLooper()).post {
            analyzerResult?.success(mapOf(
                "name" to "barcode",
                "data" to it
            ))
            analyzerResult = null
        }
    }

    private var analyzerResult: MethodChannel.Result? = null

    private val callback: BarcodeScannerCallback = { barcodes: List<Map<String, Any?>>, image: ByteArray?, width: Int?, height: Int? ->
        barcodeHandler.publishEvent(mapOf(
            "name" to "barcode",
            "data" to barcodes,
            // The image dimensions are always provided.
            // The image bytes are only non-null when `returnImage` is true.
            "image" to mapOf(
                "bytes" to image,
                "width" to width?.toDouble(),
                "height" to height?.toDouble(),
            )
        ))
    }

    private val errorCallback: BarcodeScannerErrorCallback = {error: String ->
        barcodeHandler.publishError(BarcodeScannerErrorCodes.BARCODE_ERROR, error, null)
    }

    private var methodChannel: MethodChannel? = null

    private var barcodeScanner: BarcodeScanner? = null

    private val torchStateCallback: TorchStateCallback = {state: Int ->
        // Off = 0, On = 1
        barcodeHandler.publishEvent(mapOf("name" to "torchState", "data" to state))
    }

    private val zoomScaleStateCallback: ZoomScaleStateCallback = {zoomScale: Double ->
        barcodeHandler.publishEvent(mapOf("name" to "zoomScaleState", "data" to zoomScale))
    }

    private val exposureLevelStateCallback: ExposureLevelStateCallback = {exposure: Double ->
        barcodeHandler.publishEvent(mapOf("name" to "exposureLevelState", "data" to exposure))
    }

    init {
        methodChannel = MethodChannel(binaryMessenger,
            "consoledoctor.barcode_scanner")
        methodChannel!!.setMethodCallHandler(this)
        barcodeScanner = BarcodeScanner(activity, textureRegistry, callback, errorCallback)
    }

    fun dispose(activityPluginBinding: ActivityPluginBinding) {
        methodChannel?.setMethodCallHandler(null)
        methodChannel = null
        barcodeScanner?.dispose()
        barcodeScanner = null

        val listener: RequestPermissionsResultListener? = permissions.getPermissionListener()

        if(listener != null) {
            activityPluginBinding.removeRequestPermissionsResultListener(listener)
        }
    }

    @ExperimentalGetImage
    override fun onMethodCall(call: MethodCall, result: MethodChannel.Result) {
        when (call.method) {
            "state" -> result.success(permissions.hasCameraPermission(activity))
            "request" -> permissions.requestPermission(
                activity,
                addPermissionListener,
                object: BarcodeScannerPermissions.ResultCallback {
                    override fun onResult(errorCode: String?) {
                        when(errorCode) {
                            null -> result.success(true)
                            BarcodeScannerErrorCodes.CAMERA_ACCESS_DENIED -> result.success(false)
                            BarcodeScannerErrorCodes.CAMERA_PERMISSIONS_REQUEST_ONGOING -> result.error(
                                BarcodeScannerErrorCodes.CAMERA_PERMISSIONS_REQUEST_ONGOING,
                                BarcodeScannerErrorCodes.CAMERA_PERMISSIONS_REQUEST_ONGOING_MESSAGE, null)
                            else -> result.error(
                                BarcodeScannerErrorCodes.GENERIC_ERROR, BarcodeScannerErrorCodes.GENERIC_ERROR_MESSAGE, null)
                        }
                    }
                })
            "start" -> start(call, result)
            "stop" -> stop(result)
            "toggleTorch" -> toggleTorch(result)
            "analyzeImage" -> analyzeImage(call, result)
            "setScale" -> setScale(call, result)
            "resetScale" -> resetScale(result)
            "setExposure" -> setExposure(call, result)
            "resetExposure" -> resetExposure(result)
            "updateScanWindow" -> updateScanWindow(call, result)
            "setTryInverted" -> setTryInverted(call, result)
            "setDetectionSpeed" -> setDetectionSpeed(call, result)
            else -> result.notImplemented()
        }
    }

    @ExperimentalGetImage
    private fun start(call: MethodCall, result: MethodChannel.Result) {
        val torch: Boolean = call.argument<Boolean>("torch") ?: false
        val facing: Int = call.argument<Int>("facing") ?: 0
        val formats: List<Int>? = call.argument<List<Int>>("formats")
        val returnImage: Boolean = call.argument<Boolean>("returnImage") ?: false
        val tryInverted: Boolean = call.argument<Boolean>("tryInverted") ?: false
        val speed: Int = call.argument<Int>("speed") ?: 1
        val timeout: Int = call.argument<Int>("timeout") ?: 250
        val cameraResolutionValues: List<Int>? = call.argument<List<Int>>("cameraResolution")
        val cameraResolution: Size? = if (cameraResolutionValues != null) {
            Size(cameraResolutionValues[0], cameraResolutionValues[1])
        } else {
            null
        }

        val barcodeScannerOptions: BarcodeScannerOptions? = buildBarcodeScannerOptions(formats)

        val position =
            if (facing == 0) CameraSelector.DEFAULT_FRONT_CAMERA else CameraSelector.DEFAULT_BACK_CAMERA

        barcodeScanner!!.start(
            barcodeScannerOptions,
            returnImage,
            position,
            torch,
            tryInverted,
            DetectionSpeed.fromInt(speed),
            torchStateCallback,
            zoomScaleStateCallback,
            exposureLevelStateCallback,
            barcodeScannerStartedCallback = {
                Handler(Looper.getMainLooper()).post {
                    result.success(mapOf(
                        "textureId" to it.id,
                        "size" to mapOf("width" to it.width, "height" to it.height),
                        "currentTorchState" to it.currentTorchState,
                        "numberOfCameras" to it.numberOfCameras,
                        "minExposureLevel" to it.minExposure,
                        "maxExposureLevel" to it.maxExposure,
                        "maxZoomLevel" to it.maxZoom
                    ))
                }
            },
            barcodeScannerErrorCallback = {
                Handler(Looper.getMainLooper()).post {
                    when (it) {
                        is AlreadyStarted -> {
                            result.error(
                                BarcodeScannerErrorCodes.ALREADY_STARTED_ERROR,
                                BarcodeScannerErrorCodes.ALREADY_STARTED_ERROR_MESSAGE,
                                null
                            )
                        }
                        is CameraError -> {
                            result.error(
                                BarcodeScannerErrorCodes.CAMERA_ERROR,
                                BarcodeScannerErrorCodes.CAMERA_ERROR_MESSAGE,
                                null
                            )
                        }
                        is NoCamera -> {
                            result.error(
                                BarcodeScannerErrorCodes.NO_CAMERA_ERROR,
                                BarcodeScannerErrorCodes.NO_CAMERA_ERROR_MESSAGE,
                                null
                            )
                        }
                        else -> {
                            result.error(
                                BarcodeScannerErrorCodes.GENERIC_ERROR,
                                BarcodeScannerErrorCodes.GENERIC_ERROR_MESSAGE,
                                null
                            )
                        }
                    }
                }
            },
            timeout.toLong(),
            cameraResolution
        )
    }

    private fun stop(result: MethodChannel.Result) {
        try {
            barcodeScanner!!.stop()
            result.success(null)
        } catch (e: AlreadyStopped) {
            result.success(null)
        }
    }

    private fun analyzeImage(call: MethodCall, result: MethodChannel.Result) {
        analyzerResult = result

        val formats: List<Int>? = call.argument<List<Int>>("formats")
        val filePath: String = call.argument<String>("filePath")!!

        barcodeScanner!!.analyzeImage(
            Uri.fromFile(File(filePath)),
            buildBarcodeScannerOptions(formats),
            analyzeImageSuccessCallback,
            analyzeImageErrorCallback)
    }

    private fun toggleTorch(result: MethodChannel.Result) {
        barcodeScanner?.toggleTorch()
        result.success(null)
    }

    private fun setScale(call: MethodCall, result: MethodChannel.Result) {
        try {
            barcodeScanner!!.setScale(call.arguments as Double)
            result.success(null)
        } catch (e: ZoomWhenStopped) {
            result.error(
                BarcodeScannerErrorCodes.SET_SCALE_WHEN_STOPPED_ERROR, BarcodeScannerErrorCodes.SET_SCALE_WHEN_STOPPED_ERROR_MESSAGE, null)
        } catch (e: ZoomNotInRange) {
            result.error(
                BarcodeScannerErrorCodes.GENERIC_ERROR, BarcodeScannerErrorCodes.INVALID_ZOOM_SCALE_ERROR_MESSAGE, null)
        }
    }

    private fun resetScale(result: MethodChannel.Result) {
        try {
            barcodeScanner!!.resetScale()
            result.success(null)
        } catch (e: ZoomWhenStopped) {
            result.error(
                BarcodeScannerErrorCodes.SET_SCALE_WHEN_STOPPED_ERROR, BarcodeScannerErrorCodes.SET_SCALE_WHEN_STOPPED_ERROR_MESSAGE, null)
        }
    }

    private fun setExposure(call: MethodCall, result: MethodChannel.Result) {
        try {
            barcodeScanner!!.setExposure(call.arguments as Double)
            result.success(null)
        } catch (e: ExposureWhenStopped) {
            result.error(
                BarcodeScannerErrorCodes.SET_EXPOSURE_WHEN_STOPPED_ERROR, BarcodeScannerErrorCodes.SET_EXPOSURE_WHEN_STOPPED_ERROR_MESSAGE, null)
        } catch (e: ExposureError) {
            result.error(
                BarcodeScannerErrorCodes.GENERIC_ERROR, BarcodeScannerErrorCodes.INVALID_EXPOSURE_LEVEL_ERROR_MESSAGE, null)
        }
    }

    private fun resetExposure(result: MethodChannel.Result) {
        try {
            barcodeScanner!!.resetExposure()
            result.success(null)
        } catch (e: ExposureWhenStopped) {
            result.error(
                BarcodeScannerErrorCodes.SET_EXPOSURE_WHEN_STOPPED_ERROR, BarcodeScannerErrorCodes.SET_SCALE_WHEN_STOPPED_ERROR_MESSAGE, null)
        }
    }

    private fun setTryInverted(call: MethodCall, result: MethodChannel.Result) {
        barcodeScanner!!.setTryInverted(call.arguments as Boolean)
        result.success(null)
    }

    private fun setDetectionSpeed(call: MethodCall, result: MethodChannel.Result) {
        barcodeScanner!!.setDetectionSpeed(DetectionSpeed.fromInt(call.arguments as Int))
        result.success(null)
    }

    private fun updateScanWindow(call: MethodCall, result: MethodChannel.Result) {
        barcodeScanner?.scanWindow = call.argument<List<Float>?>("rect")

        result.success(null)
    }

    private fun buildBarcodeScannerOptions(formats: List<Int>?): BarcodeScannerOptions? {
        if (formats == null) {
            return null
        }

        val formatsList: MutableList<Int> = mutableListOf()

        for (formatValue in formats) {
            formatsList.add(BarcodeFormats.fromRawValue(formatValue).intValue)
        }

        if (formatsList.size == 1) {
            return BarcodeScannerOptions.Builder().setBarcodeFormats(formatsList.first())
                .build()
        }

        return BarcodeScannerOptions.Builder().setBarcodeFormats(
            formatsList.first(),
            *formatsList.subList(1, formatsList.size).toIntArray()
        ).build()
    }
}
