import 'dart:async';

import 'package:consoledoctor_barcode_scanner/consoledoctor_barcode_scanner.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:consoledoctor_barcode_scanner/src/barcode_scanner_view_attributes.dart';
import 'package:consoledoctor_barcode_scanner/src/objects/start_options.dart';

/// An implementation of [BarcodeScannerPlatform] that uses method channels.
class MethodChannelBarcodeScanner extends BarcodeScannerPlatform {
  /// The name of the barcode event that is sent when a barcode is scanned.
  @visibleForTesting
  static const String kBarcodeEventName = 'barcode';

  /// The name of the error event that is sent when a barcode scan error occurs.
  @visibleForTesting
  static const String kBarcodeErrorEventName = 'BARCODE_SCANNER_BARCODE_ERROR';

  /// The method channel used to interact with the native platform.
  @visibleForTesting
  final methodChannel = const MethodChannel(
    'consoledoctor.barcode_scanner',
  );

  /// The event channel that sends back scanned barcode events.
  @visibleForTesting
  final eventChannel = const EventChannel(
    'consoledoctor.barcode_scanner.events',
  );

  Stream<Map<Object?, Object?>>? _eventsStream;

  /// Get the event stream of barcode events that come from the [eventChannel].
  Stream<Map<Object?, Object?>> get eventsStream {
    _eventsStream ??=
        eventChannel.receiveBroadcastStream().cast<Map<Object?, Object?>>();

    return _eventsStream!;
  }

  int? _textureId;

  /// Parse a [BarcodeCapture] from the given [event].
  BarcodeCapture? _parseBarcode(Map<Object?, Object?>? event) {
    if (event == null) {
      return null;
    }

    final Object? data = event['data'];

    if (data == null || data is! List<Object?>) {
      return null;
    }

    final List<Map<Object?, Object?>> barcodes =
        data.cast<Map<Object?, Object?>>();

    if (defaultTargetPlatform == TargetPlatform.android ||
        defaultTargetPlatform == TargetPlatform.iOS ||
        defaultTargetPlatform == TargetPlatform.macOS) {
      final Map<Object?, Object?>? imageData =
          event['image'] as Map<Object?, Object?>?;
      final Uint8List? image = imageData?['bytes'] as Uint8List?;
      final double? width = imageData?['width'] as double?;
      final double? height = imageData?['height'] as double?;

      return BarcodeCapture(
        raw: event,
        barcodes: barcodes.map(Barcode.fromNative).toList(),
        image: image,
        size: width == null || height == null ? Size.zero : Size(width, height),
      );
    }

    throw const BarcodeScannerException(
      errorCode: BarcodeScannerErrorCode.genericError,
      errorDetails: BarcodeScannerErrorDetails(
        message: 'Only Android, iOS and macOS are supported.',
      ),
    );
  }

  /// Parse a [BarcodeScannerBarcodeException] from the given [error] and [stackTrace], and throw it.
  ///
  /// If the error is not a [PlatformException],
  /// with [kBarcodeErrorEventName] as [PlatformException.code], the error is rethrown as-is.
  Never _parseBarcodeError(Object error, StackTrace stackTrace) {
    if (error case PlatformException(:final String code, :final String? message)
        when code == kBarcodeErrorEventName) {
      throw BarcodeScannerBarcodeException(message);
    }

    Error.throwWithStackTrace(error, stackTrace);
  }

  /// Request permission to access the camera.
  ///
  /// Throws a [BarcodeScannerException] if the permission is not granted.
  Future<void> _requestCameraPermission() async {
    try {
      final BarcodeScannerAuthorizationState authorizationState =
          BarcodeScannerAuthorizationState.fromRawValue(
        await methodChannel.invokeMethod<int>('state') ?? 0,
      );

      switch (authorizationState) {
        // Authorization was already granted, no need to request it again.
        case BarcodeScannerAuthorizationState.authorized:
          return;
        // Android does not have an undetermined authorization state.
        // So if the permission was denied, request it again.
        case BarcodeScannerAuthorizationState.denied:
        case BarcodeScannerAuthorizationState.undetermined:
          final bool permissionGranted =
              await methodChannel.invokeMethod<bool>('request') ?? false;

          if (!permissionGranted) {
            throw const BarcodeScannerException(
              errorCode: BarcodeScannerErrorCode.permissionDenied,
            );
          }
      }
    } on PlatformException catch (error) {
      // If the permission state is invalid, that is an error.
      throw BarcodeScannerException(
        errorCode: BarcodeScannerErrorCode.genericError,
        errorDetails: BarcodeScannerErrorDetails(
          code: error.code,
          details: error.details as Object?,
          message: error.message,
        ),
      );
    }
  }

  @override
  Stream<BarcodeCapture?> get barcodesStream {
    // Handle incoming barcode events.
    // The error events are transformed to `MobileScannerBarcodeException` where possible.
    return eventsStream
        .where((e) => e['name'] == kBarcodeEventName)
        .map((event) => _parseBarcode(event))
        .handleError(_parseBarcodeError);
  }

  @override
  Stream<TorchState> get torchStateStream {
    return eventsStream
        .where((event) => event['name'] == 'torchState')
        .map((event) => TorchState.fromRawValue(event['data'] as int? ?? 0));
  }

  @override
  Stream<double> get zoomScaleStateStream {
    return eventsStream
        .where((event) => event['name'] == 'zoomScaleState')
        .map((event) => event['data'] as double? ?? 0.0);
  }

  @override
  Stream<double> get exposureLevelStateStream {
    return eventsStream
        .where((event) => event['name'] == 'exposureLevelState')
        .map((event) => event['data'] as double? ?? 0.0);
  }

  @override
  Future<BarcodeCapture?> analyzeImage(
    String path, {
    List<BarcodeFormat> formats = const <BarcodeFormat>[],
  }) async {
    try {
      final Map<Object?, Object?>? result =
          await methodChannel.invokeMapMethod<Object?, Object?>(
        'analyzeImage',
        {
          'filePath': path,
          'formats': formats.isEmpty
              ? null
              : [
                  for (final BarcodeFormat format in formats)
                    if (format != BarcodeFormat.unknown) format.rawValue,
                ],
        },
      );

      return _parseBarcode(result);
    } on PlatformException catch (error) {
      // Handle any errors from analyze image requests.
      if (error.code == kBarcodeErrorEventName) {
        throw BarcodeScannerBarcodeException(error.message);
      }

      return null;
    }
  }

  @override
  Widget buildCameraView() {
    if (_textureId == null) {
      return const SizedBox();
    }

    return Texture(textureId: _textureId!);
  }

  @override
  Future<void> resetZoomScale() async {
    await methodChannel.invokeMethod<void>('resetScale');
  }

  @override
  Future<void> resetExposureLevel() async {
    await methodChannel.invokeMethod<void>('resetExposure');
  }

  @override
  Future<void> setZoomScale(double zoomScale) async {
    await methodChannel.invokeMethod<void>('setScale', zoomScale);
  }

  @override
  Future<void> setExposureLevel(double exposureLevel) async {
    await methodChannel.invokeMethod<void>('setExposure', exposureLevel);
  }

  @override
  Future<void> setDetectionSpeed(DetectionSpeed detection) async {
    await methodChannel.invokeMethod<void>('setDetectionSpeed', detection.rawValue);
  }

  @override
  Future<BarcodeScannerViewAttributes> start(StartOptions startOptions) async {
    if (_textureId != null) {
      throw const BarcodeScannerException(
        errorCode: BarcodeScannerErrorCode.controllerAlreadyInitialized,
        errorDetails: BarcodeScannerErrorDetails(
          message: 'The scanner was already started.',
        ),
      );
    }

    await _requestCameraPermission();

    Map<String, Object?>? startResult;

    try {
      startResult = await methodChannel.invokeMapMethod<String, Object?>(
        'start',
        startOptions.toMap(),
      );
    } on PlatformException catch (error) {
      throw BarcodeScannerException(
        errorCode: BarcodeScannerErrorCode.fromPlatformException(error),
        errorDetails: BarcodeScannerErrorDetails(
          code: error.code,
          details: error.details as Object?,
          message: error.message,
        ),
      );
    }

    if (startResult == null) {
      throw const BarcodeScannerException(
        errorCode: BarcodeScannerErrorCode.genericError,
        errorDetails: BarcodeScannerErrorDetails(
          message: 'The start method did not return a view configuration.',
        ),
      );
    }

    final int? textureId = startResult['textureId'] as int?;

    if (textureId == null) {
      throw const BarcodeScannerException(
        errorCode: BarcodeScannerErrorCode.genericError,
        errorDetails: BarcodeScannerErrorDetails(
          message: 'The start method did not return a texture id.',
        ),
      );
    }

    _textureId = textureId;

    final int? numberOfCameras = startResult['numberOfCameras'] as int?;
    final TorchState currentTorchState = TorchState.fromRawValue(
      startResult['currentTorchState'] as int? ?? -1,
    );

    final Size size;

    if (startResult['size']
        case {'width': final double width, 'height': final double height}) {
      size = Size(width, height);
    } else {
      size = Size.zero;
    }

    final double? minExposure = startResult['minExposureLevel'] as double?;
    final double? maxExposure = startResult['maxExposureLevel'] as double?;
    final double? maxZoom = startResult['maxZoomLevel'] as double?;

    return BarcodeScannerViewAttributes(
      currentTorchMode: currentTorchState,
      numberOfCameras: numberOfCameras,
      size: size,
      minExposure: minExposure ?? 0.0,
      maxExposure: maxExposure ?? 0.0,
      maxZoom: maxZoom ?? 1.0
    );
  }

  @override
  Future<void> stop() async {
    if (_textureId == null) {
      return;
    }

    _textureId = null;

    await methodChannel.invokeMethod<void>('stop');
  }

  @override
  Future<void> toggleTorch() async {
    await methodChannel.invokeMethod<void>('toggleTorch');
  }

  @override
  Future<void> shouldTryInverted(bool shouldInvert) async {
    await methodChannel.invokeMethod<void>('setTryInverted', shouldInvert);
  }

  @override
  Future<void> updateScanWindow(Rect? window) async {
    if (_textureId == null) {
      return;
    }

    List<double>? points;

    if (window != null) {
      points = [window.left, window.top, window.right, window.bottom];
    }

    await methodChannel.invokeMethod<void>(
      'updateScanWindow',
      {'rect': points},
    );
  }

  @override
  Future<void> dispose() async {
    await stop();
  }
}
